from torchnlp.text_encoders.static_tokenizer_encoder import StaticTokenizerEncoder


class IdentityEncoder(StaticTokenizerEncoder):
    """ No tokenization for example: 'Hi There' => ['Hi There'] """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('IdentityEncoder defines a identity tokenize s -> [s]')
        if 'append_eos' not in kwargs:
            kwargs['append_eos'] = False  # Default to not appending EOS
        super().__init__(*args, **kwargs, tokenize=(lambda s: [s]))

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return ''.join(tokens)



from torch.utils.data.sampler import Sampler


class SortedSampler(Sampler):
    """Samples elements sequentially, always in the same order.

    Arguments:
        data_source (datasets.Dataset): dataset to sample from
        sort_key (callable): specifies a function of one argument that is used to extract a
          comparison key from each list element
    """

    def __init__(self, data_source, sort_key):
        super().__init__(data_source)
        self.data_source = data_source
        self.sort_key = sort_key
        zip = [(i, self.sort_key(row)) for i, row in enumerate(self.data_source)]
        zip = sorted(zip, key=lambda r: r[1])
        self.sorted_indexes = [item[0] for item in zip]

    def __iter__(self):
        return iter(self.sorted_indexes)

    def __len__(self):
        return len(self.data_source)



from torchnlp.text_encoders.static_tokenizer_encoder import StaticTokenizerEncoder


class WordEncoder(StaticTokenizerEncoder):
    """ Split a string by spaces """

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return ' '.join(tokens)



import random

from torch.utils.data.sampler import BatchSampler


class RandomBatchSampler(BatchSampler):

    def __iter__(self):
        batches = list(super().__iter__())
        random.shuffle(batches)
        return iter(batches)



from collections import Counter

import torch

from torchnlp.text_encoders.reserved_tokens import EOS_INDEX
from torchnlp.text_encoders.reserved_tokens import UNKNOWN_INDEX
from torchnlp.text_encoders.reserved_tokens import RESERVED_ITOS
from torchnlp.text_encoders.reserved_tokens import RESERVED_STOI
from torchnlp.text_encoders.text_encoders import TextEncoder


class StaticTokenizerEncoder(TextEncoder):
    """ Encoder where the tokenizer is a static callable. """

    def __init__(self,
                 sample,
                 min_occurrences=1,
                 append_eos=False,
                 lower=True,
                 tokenize=(lambda s: s.split())):
        """
        Given a sample, build the dictionary for the word encoder.

        Args:
            sample (list of strings): sample of data to build dictionary on
            min_occurrences (int): minimum number of occurrences for a token to be added to
              dictionary
            append_eos (bool): if to append EOS token onto the end to the encoded vector
            lower (bool): if to lower the text
            tokenize (callable): callable to tokenize a string
        """
        if not isinstance(sample, list):
            raise TypeError('Sample needs to be a list of strings.')

        self.lower = lower
        self.tokenize = tokenize
        self.append_eos = append_eos
        self.tokens = Counter()

        for text in sample:
            self.tokens.update(self._preprocess(text))

        self.stoi = RESERVED_STOI.copy()
        self.itos = RESERVED_ITOS[:]
        for token, count in self.tokens.items():
            if count >= min_occurrences:
                self.itos.append(token)
                self.stoi[token] = len(self.itos) - 1

    @property
    def vocab(self):
        """ Return a list of tokens """
        return self.itos

    def _preprocess(self, text):
        """ Preprocess text before encoding as a tensor. """
        if self.lower:
            text = text.lower()
        text = text.rstrip('\n')
        if self.tokenize:
            text = self.tokenize(text)
        return text

    def encode(self, text):
        """ Encode text into a vector """
        text = self._preprocess(text)
        vector = [self.stoi.get(token, UNKNOWN_INDEX) for token in text]
        if self.append_eos:
            vector.append(EOS_INDEX)
        return torch.LongTensor(vector)

    def decode(self, tensor):
        """ Decode vector into text, not guaranteed to be reversable """
        tokens = [self.itos[index] for index in tensor]
        return ' '.join(tokens)



import torch
import torch.nn as nn

# TODO: Credit IBM for this snipit


class Attention(nn.Module):
    r"""
    Applies an attention mechanism on the output features from the decoder.
    Attributes:
        linear_out (torch.nn.Linear): applies a linear transformation to the incoming data:
            :math:`y = Ax + b`.
    """

    def __init__(self, dimensions, attention_type='general'):
        """
        Args:
            dimensions (int): The number of expected features in the output
        """
        super(Attention, self).__init__()

        self.attention_type = attention_type
        assert (self.attention_type in ["dot", "general"]), "Invalid attention type selected."

        if self.attention_type == 'general':
            self.linear_in = nn.Linear(dimensions, dimensions, bias=False)
        self.linear_out = nn.Linear(dimensions * 2, dimensions, bias=False)
        self.softmax = nn.Softmax(dim=0)
        self.tanh = nn.Tanh()

    def forward(self, input_, context):
        """
        Args:
            input_ (torch.FloatTensor [batch_size, output_len, dimensions]): the attention input.
            context (torch.FloatTensor [batch_size, input_len, dimensions]): tensor containing
                features of the encoded input sequence.
        Returns:
            output (torch.LongTensor [batch_size, output_len, dimensions]): tensor containing the
                attended output features.
            attention_weights (torch.FloatTensor [batch_size, output_len, input_len]): tensor
                containing attention weights.
        """
        batch_size, output_len, dimensions = input_.size()
        input_len = context.size(1)

        if self.attention_type == "general":
            input_ = input_.view(batch_size * output_len, dimensions)
            input_ = self.linear_in(input_)
            input_ = input_.view(batch_size, output_len, dimensions)

        # TODO: Include mask on PADDING_INDEX?

        # (batch_size, output_len, dimensions) * (batch_size, input_len, dimensions) ->
        # (batch_size, output_len, input_len)
        attention_scores = torch.bmm(input_, context.transpose(1, 2).contiguous())

        # Compute weights across every context sequence
        attention_scores = attention_scores.view(batch_size * output_len, input_len)
        attention_weights = self.softmax(attention_scores)
        attention_weights = attention_weights.view(batch_size, output_len, input_len)

        # (batch_size, output_len, input_len) * (batch_size, input_len, dimensions) ->
        # (batch_size, output_len, dimensions)
        mix = torch.bmm(attention_weights, context)

        # concat -> (batch_size * output_len, 2*dimensions)
        combined = torch.cat((mix, input_), dim=2)
        combined = combined.view(batch_size * output_len, 2 * dimensions)

        # Apply linear_out on every 2nd dimension of concat
        # output -> (batch_size, output_len, dimensions)
        output = self.linear_out(combined).view(batch_size, output_len, dimensions)
        output = self.tanh(output)

        return output, attention_weights



from torch.utils import data


class Dataset(data.Dataset):

    def __init__(self, rows):
        self.columns = set()
        for row in rows:
            self.columns.update(row.keys())
        self.rows = rows

    def __getitem__(self, key):
        # Given an column string return list of column values.
        if isinstance(key, str):
            if key not in self.columns:
                raise AttributeError
            return [row[key] if key in row else None for row in self.rows]
        # Given an row integer return a object of row values.
        else:
            return self.rows[key]

    def __len__(self):
        return len(self.rows)

    def __contains__(self, key):
        # Check if `self.columns` contains column
        return key in self.columns



from torchnlp.samplers.bucket_batch_sampler import BucketBatchSampler
from torchnlp.samplers.sorted_sampler import SortedSampler
from torchnlp.samplers.noisy_sorted_sampler import NoisySortedSampler
from torchnlp.samplers.random_batch_sampler import RandomBatchSampler

__all__ = ['NoisySortedSampler', 'RandomBatchSampler', 'SortedSampler', 'BucketBatchSampler']



import logging
import logging.config

import random
import torch

logger = logging.getLogger(__name__)


def resplit_datasets(dataset, other_dataset, random_seed=None, cut=None):
    """Deterministic shuffle and split algorithm.

    Given the same two datasets and the same `random_seed`, the split happens the same exact way
    every call.

    Args:
        dataset (lib.datasets.Dataset): First dataset.
        other_dataset (lib.datasets.Dataset): Another dataset.
        random_seed (int, optional): Seed to control the shuffle of both datasets.
        cut (float, optional): If defined it is the percentage of rows that first dataset gets after
            split otherwise the original proportions are kept.

    Returns:
        dataset (lib.datasets.Dataset)
        other_dataset (lib.datasets.Dataset)
    """
    # Prevent circular dependency
    from torchnlp.datasets import Dataset

    concat = dataset.rows + other_dataset.rows
    # Reference:
    # https://stackoverflow.com/questions/19306976/python-shuffling-with-a-parameter-to-get-the-same-result
    # NOTE: Shuffle the same way every call of `shuffle_datasets` where the `random_seed` is given
    random.Random(random_seed).shuffle(concat)
    if cut is None:
        return Dataset(concat[:len(dataset)]), Dataset(concat[len(dataset):])
    else:
        cut = max(min(round(len(concat) * cut), len(concat)), 0)
        return Dataset(concat[:cut]), Dataset(concat[cut:])


def torch_equals_ignore_index(tensor, tensor_other, ignore_index=None):
    """
    Compute torch.equals with the optional mask parameter.

    Args:
        ignore_index (int, optional): specifies a tensor1 index that is ignored
    Returns:
        (bool) iff target and prediction are equal
    """
    if ignore_index is not None:
        assert tensor.size() == tensor_other.size()
        mask_arr = tensor.ne(ignore_index)
        tensor = tensor.masked_select(mask_arr)
        tensor_other = tensor_other.masked_select(mask_arr)

    return torch.equal(tensor, tensor_other)


def reporthook(t):
    """https://github.com/tqdm/tqdm"""
    last_b = [0]

    def inner(b=1, bsize=1, tsize=None):
        """
        b: int, optionala
        Number of blocks just transferred [default: 1].
        bsize: int, optional
        Size of each block (in tqdm units) [default: 1].
        tsize: int, optional
        Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            t.total = tsize
        t.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return inner



import logging
import os
import tarfile
import urllib.request

import pandas as pd
from tqdm import tqdm

from torchnlp.datasets.dataset import Dataset
from torchnlp.utils import reporthook

logger = logging.getLogger(__name__)


def _download_simple_qa_dataset(directory, some_file='annotated_fb_data_train.txt'):
    """ Download the Simple Questions dataset into `directory`

    Args:
        directory (str)
        some_file (str): Used to make sure Simple Questions was downloaded and extracted.
     """
    if os.path.isdir(directory) and os.path.isfile(os.path.join(directory, some_file)):
        # Already downloaded
        return

    if not os.path.isdir(directory):
        os.makedirs(directory)

    url = 'https://www.dropbox.com/s/tohrsllcfy7rch4/SimpleQuestions_v2.tgz?raw=1'
    filename = os.path.join(directory, 'SimpleQuestions_v2.tgz')
    logger.info('Downloading Simple Questions...')
    with tqdm() as t:  # all optional kwargs
        urllib.request.urlretrieve(url, filename=filename, reporthook=reporthook(t))
    logger.info('Extracting Simple Questions...')
    tarfile_ = tarfile.open(filename, mode='r')
    for member in tarfile_.getmembers():
        if member.isreg() and '._' != os.path.split(
                member.name)[1][:2]:  # skip if the TarInfo is not files and not hidden file
            member.name = os.path.normpath(member.name)
            # remove the root folder SimpleQuestions_v2
            member.name = os.path.join(*member.name.split('/')[1:])
            tarfile_.extract(member=member, path=directory)
    tarfile_.close()


def simple_qa_dataset(directory='data/simple_qa',
                      train=False,
                      dev=False,
                      test=False,
                      train_filename='annotated_fb_data_train.txt',
                      dev_filename='annotated_fb_data_valid.txt',
                      test_filename='annotated_fb_data_test.txt'):
    """
    Dataset introduced by this paper:
    https://research.fb.com/publications/large-scale-simple-question-answering-with-memory-networks/

    Example:
        First row from the development `annotated_fb_data_valid` dataset::

            subject: Who was the trump ocean club international hotel and tower named after
            relation: www.freebase.com/symbols/namesake/named_after
            object: www.freebase.com/m/0cqt90
            question: www.freebase.com/m/0f3xg_
    """
    _download_simple_qa_dataset(directory, train_filename)

    ret = []
    datasets = [(train, train_filename), (dev, dev_filename), (test, test_filename)]
    for is_requested, filename in datasets:
        if not is_requested:
            continue
        full_path = os.path.join(directory, filename)
        data = pd.read_table(
            full_path, header=None, names=['subject', 'relation', 'object', 'question'])
        rows = []
        for _, row in data.iterrows():
            rows.append({
                'question': row['question'],
                'relation': row['relation'],
                'object': row['object'],
                'subject': row['subject'],
            })
        ret.append(Dataset(rows))

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)



import random

from torch.utils.data.sampler import Sampler


class NoisySortedSampler(Sampler):
    """Samples elements sequentially with noise.

    Reference and inspiration:
    https://github.com/allenai/allennlp/blob/e125a490b71b21e914af01e70e9b00b165d64dcd/allennlp/data/iterators/bucket_iterator.py

    Arguments:
        data_source (datasets.Dataset): dataset to sample from
        sort_key (callable): specifies a function of one argument that is used to extract a
          comparison key from each list element
    """

    def __init__(self, data_source, sort_key, sort_key_noise=0.25):
        super().__init__(data_source)
        self.data_source = data_source
        self.sort_key = sort_key
        zip_ = []
        for i, row in enumerate(self.data_source):
            value = self.sort_key(row)
            noise_value = value * sort_key_noise
            noise = random.uniform(-noise_value, noise_value)
            value = noise + value
            zip_.append(tuple([i, value]))
        zip_ = sorted(zip_, key=lambda r: r[1])
        self.sorted_indexes = [item[0] for item in zip_]

    def __iter__(self):
        return iter(self.sorted_indexes)

    def __len__(self):
        return len(self.data_source)



from torchnlp.datasets.dataset import Dataset


def zero_to_zero_dataset(train=False,
                         dev=False,
                         test=False,
                         train_rows=256,
                         dev_rows=64,
                         test_rows=64):
    """
    Used for quick integration testing.

    Sample Data:
        Input: 0
        Output: 0
    """
    ret = []
    for is_requested, n_rows in [(train, train_rows), (dev, dev_rows), (test, test_rows)]:
        if not is_requested:
            continue
        rows = [{'source': str(0), 'target': str(0)} for i in range(n_rows)]
        ret.append(Dataset(rows))

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)



class TextEncoder(object):

    def __init__(self):
        raise NotImplementedError

    def encode(self, string):
        """ Given a string encode it into a tensor """
        raise NotImplementedError

    def decode(self, tensor):
        """ Given a tensor decode it into a string """
        raise NotImplementedError

    @property
    def vocab_size(self):
        """ Return the size of the vocab used to encode the text """
        return len(self.vocab)

    @property
    def vocab(self):
        """ Return an array of the vocab such that index matches the token in encode """
        return NotImplementedError



from torchnlp.text_encoders.static_tokenizer_encoder import StaticTokenizerEncoder
from torchnlp.text_encoders.reserved_tokens import RESERVED_STOI
from torchnlp.text_encoders.reserved_tokens import UNKNOWN_TOKEN


class DelimiterEncoder(StaticTokenizerEncoder):
    """ Encode by splitting up by a delimiter """

    def __init__(self, delimiter, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('CharacterEncoder defines a tokenize callable per character')
        self.delimiter = delimiter
        super().__init__(*args, **kwargs, tokenize=(lambda s: s.split(delimiter)))

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]

        # NOTE: Join reserved tokens like `PAD_TOKEN` and `EOS_TOKEN` with '' instead of delimiter
        # for aesthetic reasons at the end of the text phrase
        reserved = []
        while len(tokens) > 0 and tokens[-1] in RESERVED_STOI and tokens[-1] != UNKNOWN_TOKEN:
            reserved.insert(0, tokens.pop())

        return self.delimiter.join(tokens) + ''.join(reserved)



from __future__ import unicode_literals

import array
import io
import logging
import os
import zipfile

from six.moves.urllib.request import urlretrieve
from tqdm import tqdm

import six
import tarfile
import torch

from torchnlp.utils import reporthook

logger = logging.getLogger(__name__)

# TODO: Credit torchtext for this snippet


class _PretrainedEmbeddings(object):

    def __init__(self,
                 name,
                 cache='.pretrained_embeddings_cache',
                 url=None,
                 unk_init=torch.Tensor.zero_,
                 is_include=None):
        """
        Args:
            name: name of the file that contains the vectors
            cache: directory for cached vectors
            url: url for download if vectors not found in cache
            unk_init (callback): by default, initialize out-of-vocabulary word vectors
                to zero vectors; can be any function that takes in a Tensor and
                returns a Tensor of the same size
            is_include (callable): callable returns True if to include a token in memory vectors
                cache; some of these embedding files are gigantic so filtering it can cut
                down on the memory usage. We do not cache on disk if is_include is defined.

        """
        self.unk_init = unk_init
        self.is_include = is_include
        self.name = name
        self.cache(name, cache, url=url)

    def __getitem__(self, token):
        if token in self.stoi:
            return self.vectors[self.stoi[token]]
        else:
            return self.unk_init(torch.Tensor(1, self.dim))

    def __len__(self):
        return len(self.vectors)

    def __str__(self):
        return self.name

    def cache(self, name, cache, url=None):
        if os.path.isfile(name):
            path = name
            path_pt = os.path.join(cache, os.path.basename(name)) + '.pt'
        else:
            path = os.path.join(cache, name)
            path_pt = path + '.pt'

        print(path)

        if not os.path.isfile(path_pt) or self.is_include is not None:
            if not os.path.isfile(path) and url:
                logger.info('Downloading vectors from {}'.format(url))
                if not os.path.exists(cache):
                    os.makedirs(cache)
                dest = os.path.join(cache, os.path.basename(url))
                print(dest)
                if not os.path.isfile(dest):
                    with tqdm(unit='B', unit_scale=True, miniters=1, desc=dest) as t:
                        urlretrieve(url, dest, reporthook=reporthook(t))
                logger.info('Extracting vectors into {}'.format(cache))
                ext = os.path.splitext(dest)[1][1:]
                print(ext)
                if ext == 'zip':
                    with zipfile.ZipFile(dest, "r") as zf:
                        zf.extractall(cache)
                elif ext == 'gz':
                    with tarfile.open(dest, 'r:gz') as tar:
                        tar.extractall(path=cache)
            if not os.path.isfile(path):
                raise RuntimeError('no vectors found at {}'.format(path))

            # str call is necessary for Python 2/3 compatibility, since
            # argument must be Python 2 str (Python 3 bytes) or
            # Python 3 str (Python 2 unicode)
            itos, vectors, dim = [], array.array(str('d')), None

            # Try to read the whole file with utf-8 encoding.
            binary_lines = False
            try:
                with io.open(path, encoding="utf8") as f:
                    lines = [line for line in f]
            # If there are malformed lines, read in binary mode
            # and manually decode each word from utf-8
            except:
                logger.warning("Could not read {} as UTF8 file, "
                               "reading file as bytes and skipping "
                               "words with malformed UTF8.".format(path))
                with open(path, 'rb') as f:
                    lines = [line for line in f]
                binary_lines = True

            logger.info("Loading vectors from {}".format(path))
            for line in tqdm(lines, total=len(lines)):
                # Explicitly splitting on " " is important, so we don't
                # get rid of Unicode non-breaking spaces in the vectors.
                entries = line.rstrip().split(b" " if binary_lines else " ")

                word, entries = entries[0], entries[1:]
                if dim is None and len(entries) > 1:
                    dim = len(entries)
                elif len(entries) == 1:
                    logger.warning("Skipping token {} with 1-dimensional "
                                   "vector {}; likely a header".format(word, entries))
                    continue
                elif dim != len(entries):
                    raise RuntimeError("Vector for token {} has {} dimensions, but previously "
                                       "read vectors have {} dimensions. All vectors must have "
                                       "the same number of dimensions.".format(
                                           word, len(entries), dim))

                if binary_lines:
                    try:
                        if isinstance(word, six.binary_type):
                            word = word.decode('utf-8')
                    except:
                        logger.info("Skipping non-UTF8 token {}".format(repr(word)))
                        continue

                if self.is_include is not None and not self.is_include(word):
                    continue

                vectors.extend(float(x) for x in entries)
                itos.append(word)

            self.itos = itos
            self.stoi = {word: i for i, word in enumerate(itos)}
            self.vectors = torch.Tensor(vectors).view(-1, dim)
            self.dim = dim
            logger.info('Saving vectors to {}'.format(path_pt))
            torch.save((self.itos, self.stoi, self.vectors, self.dim), path_pt)
        else:
            logger.info('Loading vectors from {}'.format(path_pt))
            self.itos, self.stoi, self.vectors, self.dim = torch.load(path_pt)


class GloVe(_PretrainedEmbeddings):
    url = {
        '42B': 'http://nlp.stanford.edu/data/glove.42B.300d.zip',
        '840B': 'http://nlp.stanford.edu/data/glove.840B.300d.zip',
        'twitter.27B': 'http://nlp.stanford.edu/data/glove.twitter.27B.zip',
        '6B': 'http://nlp.stanford.edu/data/glove.6B.zip',
    }

    def __init__(self, name='840B', dim=300, **kwargs):
        url = self.url[name]
        name = 'glove.{}.{}d.txt'.format(name, str(dim))
        super(GloVe, self).__init__(name, url=url, **kwargs)


class FastText(_PretrainedEmbeddings):

    url_base = 'https://s3-us-west-1.amazonaws.com/fasttext-vectors/wiki.{}.vec'

    def __init__(self, language="en", **kwargs):
        url = self.url_base.format(language)
        name = os.path.basename(url)
        super(FastText, self).__init__(name, url=url, **kwargs)


class CharNGram(_PretrainedEmbeddings):

    name = 'charNgram.txt'
    url = ('http://www.logos.t.u-tokyo.ac.jp/~hassy/publications/arxiv2016jmt/'
           'jmt_pre-trained_embeddings.tar.gz')

    def __init__(self, **kwargs):
        super(CharNGram, self).__init__(self.name, url=self.url, **kwargs)

    def __getitem__(self, token):
        vector = torch.Tensor(self.dim).zero_()
        if token == "<unk>":
            return self.unk_init(vector)
        # These literals need to be coerced to unicode for Python 2 compatibility
        # when we try to join them with read ngrams from the files.
        chars = ['#BEGIN#'] + list(token) + ['#END#']
        num_vectors = 0
        for n in [2, 3, 4]:
            end = len(chars) - n + 1
            grams = [chars[i:(i + n)] for i in range(end)]
            for gram in grams:
                gram_key = '{}gram-{}'.format(n, ''.join(gram))
                if gram_key in self.stoi:
                    vector += self.vectors[self.stoi[gram_key]]
                    num_vectors += 1
        if num_vectors > 0:
            vector /= num_vectors
        else:
            vector = self.unk_init(vector)
        return vector



from torchnlp.text_encoders.static_tokenizer_encoder import StaticTokenizerEncoder


class CharacterEncoder(StaticTokenizerEncoder):
    """ Encode by splitting up by character """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('CharacterEncoder defines a tokenize callable per character')
        super().__init__(*args, **kwargs, tokenize=(lambda s: list(s)))

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return ''.join(tokens)



import spacy
from spacy.lang.en import English

from torchnlp.text_encoders.word_encoder import WordEncoder

# Use the SpacyEncoder by downloading en_core_web_sm via: `python -m spacy download en_core_web_sm`
_MODEL = spacy.load('en_core_web_sm')
tokenizer = English().Defaults.create_tokenizer(_MODEL)


class SpacyEncoder(WordEncoder):
    """ Use Moses to encode. """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('SpacyEncoder defines a tokenize callable.')

        super().__init__(*args, **kwargs, tokenize=lambda s: [w.text for w in tokenizer(s)])



from torchnlp.datasets.reverse import reverse_dataset
from torchnlp.datasets.count import count_dataset
from torchnlp.datasets.zero_to_zero import zero_to_zero_dataset
from torchnlp.datasets.simple_qa import simple_qa_dataset
from torchnlp.datasets.dataset import Dataset

__all__ = [
    'Dataset', 'simple_qa_dataset', 'zero_to_zero_dataset', 'count_dataset', 'reverse_dataset'
]



from torchnlp.text_encoders.word_encoder import WordEncoder


class MosesEncoder(WordEncoder):
    """ Use Moses to encode. """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('MosesEncoder defines a tokenize callable Moses')

        import nltk

        # Required for moses
        nltk.download('perluniprops')
        nltk.download('nonbreaking_prefixes')

        from nltk.tokenize.moses import MosesTokenizer
        from nltk.tokenize.moses import MosesDetokenizer

        self.detokenizer = MosesDetokenizer()

        super().__init__(*args, **kwargs, tokenize=MosesTokenizer().tokenize)

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return self.detokenizer.detokenize(tokens, return_str=True)



import random

from torch.utils.data.sampler import BatchSampler

from torchnlp.samplers.noisy_sorted_sampler import NoisySortedSampler


class BucketBatchSampler(BatchSampler):
    """
    Reference:
    https://github.com/allenai/allennlp/blob/e125a490b71b21e914af01e70e9b00b165d64dcd/allennlp/data/iterators/bucket_iterator.py
    https://github.com/pytorch/text/tree/master/torchtext/data/iterators/#BucketIterator

    `BucketIterator` pools together examples with a similar size length to reduce the padding
    required for each batch. `BucketIterator` typically also includes the ability to add noise to
    the pooling.

    The functionality has been replicated as a `Sampler` to be used with a
    `torch.data.utils.DataLoader`.
    """

    def __init__(self,
                 data_source,
                 sort_key,
                 batch_size,
                 sort_key_noise=0.25,
                 last_batch_first=True,
                 shuffle=True,
                 drop_last=False):
        self.last_batch_first = last_batch_first
        self.shuffle = shuffle
        super().__init__(
            NoisySortedSampler(
                data_source=data_source, sort_key=sort_key, sort_key_noise=sort_key_noise),
            batch_size, drop_last)

    def __iter__(self):
        batches = list(super().__iter__())
        if self.last_batch_first:
            last_batch = batches.pop()
        if self.shuffle:
            random.shuffle(batches)
        if self.last_batch_first:
            batches.insert(0, last_batch)
        return iter(batches)



import torch

from torchnlp.utils import torch_equals_ignore_index


def get_accuracy(targets, outputs, k=1, ignore_index=None):
    """
    Compute the accuracy of o == t {o \in outputs and t \in targets.

    Args:
      targets (list of tensors)
      outputs (list of tensors)
      ignore_index (optional, int): Specifies a target value that is ignored computing equality
    Returns:
      accuracy (float)
      number correct (int)
      total (int)
    """
    n_correct = 0.0
    for target, output in zip(targets, outputs):
        if not torch.is_tensor(target):
            target = torch.LongTensor([target])

        if not torch.is_tensor(output):
            output = torch.LongTensor([[output]])

        predictions = output.topk(k=min(k, len(output)), dim=0)[0]
        for prediction in predictions:
            if not torch.is_tensor(prediction):
                prediction = torch.LongTensor([prediction])

            if torch_equals_ignore_index(target, prediction, ignore_index=ignore_index):
                n_correct += 1
                break

    return n_correct / len(targets), n_correct, len(targets)


def get_token_accuracy(targets, outputs, ignore_index=None):
    """ Compute the token accuracy.

    Args:
      targets (list of tensors)
      outputs (list of tensors)
      ignore_index (optional, int): Specifies a target value that is ignored computing equality
    Returns:
      accuracy (float)
      number correct (int)
      total (int)
     """
    n_correct = 0.0
    n_total = 0.0
    for target, output in zip(targets, outputs):
        if not torch.is_tensor(target):
            target = torch.LongTensor([target])

        if not torch.is_tensor(output):
            output = torch.LongTensor([[output]])

        prediction = output.max(dim=0)[0].view(-1)
        if ignore_index is not None:
            mask = target.ne(ignore_index)
            n_correct += prediction.eq(target).masked_select(mask).sum()
            n_total += mask.sum()
        else:
            n_total += len(target)
            n_correct += prediction.eq(target).sum()

    return n_correct / n_total, n_correct, n_total



from torchnlp.text_encoders.word_encoder import WordEncoder


class TreebankEncoder(WordEncoder):
    """ Use Moses to encode. """

    def __init__(self, *args, **kwargs):
        if 'tokenize' in kwargs:
            raise TypeError('TreebankEncoder defines a tokenize callable TreebankWordTokenizer')

        import nltk

        # Required for moses
        nltk.download('perluniprops')
        nltk.download('nonbreaking_prefixes')

        from nltk.tokenize.treebank import TreebankWordTokenizer
        from nltk.tokenize.treebank import TreebankWordDetokenizer

        self.detokenizer = TreebankWordDetokenizer()

        super().__init__(*args, **kwargs, tokenize=TreebankWordTokenizer().tokenize)

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return self.detokenizer.detokenize(tokens)



import torch

from torchnlp.text_encoders.reserved_tokens import EOS_INDEX
from torchnlp.text_encoders.reserved_tokens import UNKNOWN_INDEX
from torchnlp.text_encoders.reserved_tokens import RESERVED_ITOS
from torchnlp.text_encoders.reserved_tokens import RESERVED_STOI
from torchnlp.text_encoders.subword_text_tokenizer import SubwordTextTokenizer
from torchnlp.text_encoders.text_encoders import TextEncoder


class SubwordEncoder(TextEncoder):
    """ Use Googles Tensor2Tensor SubwordTextTokenizer """

    def __init__(self,
                 sample,
                 append_eos=False,
                 lower=True,
                 target_vocab_size=None,
                 min_occurrences=1,
                 max_occurrences=1e3):
        """ Given a sample, build the dictionary for the word encoder.

        Args:
            sample (list of str)
            append_eos (bool)
            lower (bool)
            target_vocab_size (int): desired vocab_size to approximate
            min_occurrences (int): lower bound for the minimum token count
            max_occurrences (int): upper bound for the minimum token count
        """
        self.lower = lower
        self.append_eos = append_eos

        if self.lower:
            sample = [text.lower().rstrip('\n') for text in sample]

        if target_vocab_size is None:
            self.tokenizer = SubwordTextTokenizer()
            self.tokenizer.build_from_corpus(sample, min_count=min_occurrences)
        else:
            self.tokenizer = SubwordTextTokenizer.build_to_target_size_from_corpus(
                sample,
                target_size=target_vocab_size,
                min_val=min_occurrences,
                max_val=max_occurrences)

        self.stoi = RESERVED_STOI.copy()
        self.itos = RESERVED_ITOS[:]
        for token in self.tokenizer.vocab:
            self.itos.append(token)
            self.stoi[token] = len(self.itos) - 1

    @property
    def vocab(self):
        return self.itos

    def encode(self, text):
        if self.lower:
            text = text.lower()
        text = text.rstrip('\n')
        text = self.tokenizer.encode(text)
        vector = [self.stoi.get(token, UNKNOWN_INDEX) for token in text]
        if self.append_eos:
            vector.append(EOS_INDEX)
        return torch.LongTensor(vector)

    def decode(self, tensor):
        tokens = [self.itos[index] for index in tensor]
        return self.tokenizer.decode(tokens)



import random

from torchnlp.datasets.dataset import Dataset


def reverse_dataset(train=False,
                    dev=False,
                    test=False,
                    train_rows=10000,
                    dev_rows=1000,
                    test_rows=1000,
                    seq_max_length=10):
    """
    Used for sequence to sequence tests.

    Sample Data:
        Input: 1 2 3
        Output: 3 2 1
    """
    ret = []
    for is_requested, n_rows in [(train, train_rows), (dev, dev_rows), (test, test_rows)]:
        rows = []
        for i in range(n_rows):
            length = random.randint(1, seq_max_length)
            seq = []
            for _ in range(length):
                seq.append(str(random.randint(0, 9)))
            input_ = ' '.join(seq)
            output = ' '.join(reversed(seq))
            rows.append({'source': input_, 'target': output})

        # NOTE: Given that `random.randint` is deterministic with the same `random_seed` we need
        # to allow the random generator to create the train, dev and test dataset in order.
        # Otherwise, `reverse(train=True)` and `reverse(test=True)` would share the first 1000 rows.
        if not is_requested:
            continue

        ret.append(Dataset(rows))

    if len(ret) == 1:
        return ret[0]
    else:
        return tuple(ret)
